/** Galois type traits -*- C++ -*-
 * @file
 * @section License
 *
 * Galois, a framework to exploit amorphous data-parallelism in irregular
 * programs.
 *
 * Copyright (C) 2013, The University of Texas at Austin. All rights reserved.
 * UNIVERSITY EXPRESSLY DISCLAIMS ANY AND ALL WARRANTIES CONCERNING THIS
 * SOFTWARE AND DOCUMENTATION, INCLUDING ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR ANY PARTICULAR PURPOSE, NON-INFRINGEMENT AND WARRANTIES OF
 * PERFORMANCE, AND ANY WARRANTY THAT MIGHT OTHERWISE ARISE FROM COURSE OF
 * DEALING OR USAGE OF TRADE.  NO WARRANTY IS EITHER EXPRESS OR IMPLIED WITH
 * RESPECT TO THE USE OF THE SOFTWARE OR DOCUMENTATION. Under no circumstances
 * shall University be liable for incidental, special, indirect, direct or
 * consequential damages or loss of profits, interruption of business, or
 * related expenses which may arise from use of Software or Documentation,
 * including but not limited to those resulting from defects in Software and/or
 * Documentation, or loss or inaccuracy of data of any kind.
 *
 * @section Description
 *
 * There are two ways to declare a typetrait. First, with a typedef or other
 * valid name declaration:
 * \code
 * struct MyClass {
 *   typedef int tt_needs_parallel_break;
 *   ....
 * };
 * \endcode
 *
 * The second way is by specializing a function:
 * \code
 * namespace Galois {
 *   template<>
 *   struct needs_parallel_break<MyClass> : public boost::true_type {};
 * }
 * \endcode
 *
 * Since the compiler doesn't check the names of these traits, a good
 * programming practice is to add a <code>static_assert</code> to check if
 * everything is ok:
 * \code
 * struct MyClass {
 *   typedef int tt_needs_parallel_break;
 *   static_assert(Galois::needs_parallel_break<MyClass>::value, "Oops!");
 *   ...
 * };
 * \endcode
 *
 * @author Andrew Lenharth <andrewl@lenharth.org>
 */
#ifndef GALOIS_TYPETRAITS_H
#define GALOIS_TYPETRAITS_H

#include "Galois/Runtime/ll/CompilerSpecific.h"
#include <boost/mpl/has_xxx.hpp>

namespace Galois {

#define GALOIS_HAS_MEM_FUNC(func, name) \
  template<typename T, typename Sig> \
  struct has_##name { \
    typedef char yes[1]; \
    typedef char no[2]; \
    template<typename U, U> struct type_check; \
    template<typename W> static yes& test(type_check<Sig, &W::func>*); \
    template<typename  > static no&  test(...); \
    static const bool value = sizeof(test<T>(0)) == sizeof(yes); \
  }

#define GALOIS_HAS_MEM_FUNC_ANY(func, name) \
  template<typename T> \
  struct has_##name { \
    typedef char yes[1]; \
    typedef char no[2]; \
    template<typename U, U> struct type_check; \
    template<typename W> static yes& test(type_check<decltype(&W::func), &W::func>*); \
    template<typename  > static no&  test(...); \
    static const bool value = sizeof(test<T>(0)) == sizeof(yes); \
  }

#define GALOIS_HAS_MEM_TYPE(func, name) \
  template<typename T> \
  struct has_##name { \
    typedef char yes[1]; \
    typedef char no[2]; \
    template<typename W> static yes& test(typename W::func*); \
    template<typename  > static no&  test(...); \
    static const bool value = sizeof(test<T>(0)) == sizeof(yes); \
  }

GALOIS_HAS_MEM_FUNC(galoisDeterministicParallelBreak, tf_deterministic_parallel_break);
/**
 * Indicates the operator has a member function that allows a {@link Galois::for_each}
 * loop to be exited deterministically.
 *
 * The function has the following signature:
 * \code
 *  struct T {
 *    bool galoisDeterministicParallelBreak() {
 *      // returns true if loop should end
 *    }
 *  };
 *  \endcode
 *
 * This function will be periodically called by the deterministic scheduler.
 * If it returns true, the loop ends as if calling {@link
 * UserContext::breakLoop}, but unlike that function, these breaks are
 * deterministic.
 */
template<typename T>
struct has_deterministic_parallel_break : public has_tf_deterministic_parallel_break<T, bool(T::*)()> {};

GALOIS_HAS_MEM_FUNC_ANY(galoisDeterministicId, tf_deterministic_id);
/**
 * Indicates the operator has a member function that optimizes the generation
 * of unique ids for active elements. This function should be thread-safe.
 *
 * The type conforms to the following:
 * \code
 *  struct T {
 *    uintptr_t galoisDeterministicId(const A& item) const { 
 *      // returns a unique identifier for item
 *    }
 *  };
 * \endcode
 */
template<typename T>
struct has_deterministic_id : public has_tf_deterministic_id<T> {};

GALOIS_HAS_MEM_TYPE(GaloisDeterministicLocalState, tf_deterministic_local_state);
/**
 * Indicates the operator has a member type that encapsulates state that is passed between 
 * the suspension and resumpsion of an operator during deterministic scheduling.
 *
 * The type conforms to the following:
 * \code
 *  struct T {
 *    struct GaloisDeteministicLocalState {
 *      int x, y, z; // Local state
 *      GaloisDeterministicLocalState(T& self, Galois::PerIterAllocTy& alloc) {
 *        // initialize local state
 *      }
 *    };
 *
 *    void operator()(const A& item, Galois::UserContext<A>&) { 
 *      // An example of using local state
 *      typedef GaloisDeterministicLocalState LS;
 *      bool used;
 *      LS* p = (LS*) ctx.getLocalState(used);
 *      if (used) {
 *        // operator is being resumed; use p
 *      } else {
 *        // operator hasn't been suspended yet; execute normally
 *        // save state into p to be used when operator resumes
 *      }
 *    }
 *  };
 * \endcode 
 */
template<typename T>
struct has_deterministic_local_state : public has_tf_deterministic_local_state<T> {};

/**
 * Indicates the operator may request the parallel loop to be suspended and a
 * given function run in serial
 */
BOOST_MPL_HAS_XXX_TRAIT_DEF(tt_needs_parallel_break)
template<typename T>
struct needs_parallel_break : public has_tt_needs_parallel_break<T> {};

/**
 * Indicates the operator does not generate new work and push it on the worklist
 */
BOOST_MPL_HAS_XXX_TRAIT_DEF(tt_does_not_need_push)
template<typename T>
struct does_not_need_push : public has_tt_does_not_need_push<T> {};

/**
 * Indicates the operator may request the access to a per-iteration 
 * allocator
 */
BOOST_MPL_HAS_XXX_TRAIT_DEF(tt_needs_per_iter_alloc)
template<typename T>
struct needs_per_iter_alloc : public has_tt_needs_per_iter_alloc<T> {};

/**
 * Indicates the operator doesn't need its execution stats recorded
 */
BOOST_MPL_HAS_XXX_TRAIT_DEF(tt_does_not_need_stats)
template<typename T>
struct does_not_need_stats : public has_tt_does_not_need_stats<T> {};

/**
 * Indicates the operator doesn't need abort support
 */
BOOST_MPL_HAS_XXX_TRAIT_DEF(tt_does_not_need_aborts)
template<typename T>
struct does_not_need_aborts : public has_tt_does_not_need_aborts<T> {};

/**
 * Indicates that the neighborhood set does not change through out i.e. is not
 * dependent on computed values. Examples of such fixed neighborhood is e.g. the 
 * neighborhood being all the neighbors of a node in the input graph, while the
 * counter example is the neighborhood being some of the neighbors based on
 * some predicate. 
 */
BOOST_MPL_HAS_XXX_TRAIT_DEF(tt_has_fixed_neighborhood)
template <typename T>
struct has_fixed_neighborhood: public has_tt_has_fixed_neighborhood<T> {};

/**
 * Temporary type trait for pre-C++11 compilers, which don't support exact
 * std::is_trivially_constructible. 
 */
BOOST_MPL_HAS_XXX_TRAIT_DEF(tt_has_known_trivial_constructor)
template <typename T>
struct has_known_trivial_constructor: public has_tt_has_known_trivial_constructor<T> { };

}
#endif
